@@ -177,6 +177,10 @@ impl PooledBuf {
         &self.0
     }
 
+    pub fn drain_start(&mut self, start: usize) {
+        self.0.drain(..start);
+    }
+
     pub fn write(&mut self) -> &mut Vec<u8> {
         let buf = &mut self.0;
         buf.clear();
@@ -211,3 +215,31 @@ impl BufPool {
         self.pool.push(ManuallyDrop::new(buf));
     }
 }
+
+pub fn drop_event(buf_pool: &mut BufPool, event: PlatformEvent) {
+    match event {
+        PlatformEvent::ConnectionOutput { buf, .. } | PlatformEvent::ProcessOutput { buf, .. } => {
+            buf_pool.release(buf);
+        }
+        PlatformEvent::Idle
+        | PlatformEvent::ConnectionOpen { .. }
+        | PlatformEvent::ConnectionClose { .. }
+        | PlatformEvent::ProcessSpawned { .. }
+        | PlatformEvent::ProcessExit { .. } => (),
+    }
+}
+
+pub fn drop_request(buf_pool: &mut BufPool, request: PlatformRequest) {
+    match request {
+        PlatformRequest::WriteToClient { buf, .. }
+        | PlatformRequest::WriteToProcess { buf, .. } => {
+            buf_pool.release(buf);
+        }
+        PlatformRequest::Quit
+        | PlatformRequest::Redraw
+        | PlatformRequest::CloseClient { .. }
+        | PlatformRequest::SpawnProcess { .. }
+        | PlatformRequest::CloseProcessInput { .. }
+        | PlatformRequest::KillProcess { .. } => (),
+    }
+}
