@@ -1,4 +1,5 @@
 use std::{
+    collections::VecDeque,
     io,
     os::unix::{
         io::{AsRawFd, RawFd},
@@ -8,15 +9,21 @@ use std::{
 };
 
 use crate::{
-    application::{ApplicationConfig, ClientApplication, ServerApplication},
+    application::{
+        ApplicationConfig, ClientApplication, ServerApplication, CLIENT_CONNECTION_BUFFER_LEN,
+        CLIENT_STDIN_BUFFER_LEN, SERVER_CONNECTION_BUFFER_LEN, SERVER_IDLE_DURATION,
+    },
     client::ClientHandle,
-    platform::{Key, PlatformEvent, PlatformProcessHandle, PlatformRequest},
+    platform::{
+        drop_request, Key, PlatformEvent, PlatformProcessHandle, PlatformRequest, PooledBuf,
+    },
     Args,
 };
 
 mod unix_utils;
 use unix_utils::{
-    is_pipped, read, read_from_connection, run, suspend_process, write_all_bytes, Process, Terminal,
+    is_pipped, read, read_from_connection, run, suspend_process, write_all_bytes,
+    write_to_connection, Process, Terminal,
 };
 
 const MAX_CLIENT_COUNT: usize = 20;
@@ -33,9 +40,15 @@ fn errno() -> libc::c_int {
     unsafe { *libc::__error() }
 }
 
+enum EventKind {
+    Read,
+    Write,
+}
+
 enum Event {
     Resize,
-    Fd(RawFd),
+    FdRead(RawFd),
+    FdWrite(RawFd),
 }
 impl Event {
     pub fn into_kevent(self, flags: u16, index: usize) -> libc::kevent {
@@ -48,7 +61,7 @@ impl Event {
                 data: 0,
                 udata: index as _,
             },
-            Self::Fd(fd) => libc::kevent {
+            Self::FdRead(fd) => libc::kevent {
                 ident: fd as _,
                 filter: libc::EVFILT_READ,
                 flags,
@@ -56,6 +69,14 @@ impl Event {
                 data: 0,
                 udata: index as _,
             },
+            Self::FdWrite(fd) => libc::kevent {
+                ident: fd as _,
+                filter: libc::EVFILT_WRITE,
+                flags,
+                fflags: 0,
+                data: 0,
+                udata: index as _,
+            },
         }
     }
 }
@@ -63,6 +84,7 @@ impl Event {
 struct TriggeredEvent {
     pub index: usize,
     pub data: isize,
+    pub kind: EventKind,
 }
 
 struct KqueueEvents([libc::kevent; MAX_TRIGGERED_EVENT_COUNT]);
@@ -94,8 +116,8 @@ impl Kqueue {
         Self(fd)
     }
 
-    pub fn add(&self, event: Event, index: usize) {
-        let event = event.into_kevent(libc::EV_ADD, index);
+    pub fn add(&self, event: Event, index: usize, extra_flags: u16) {
+        let event = event.into_kevent(libc::EV_ADD | extra_flags, index);
         if !modify_kqueue(self.0, &event) {
             panic!("could not add event, errno: {}", errno());
         }
@@ -148,9 +170,16 @@ impl Kqueue {
             if e.flags & libc::EV_ERROR != 0 {
                 Err(())
             } else {
+                let kind = match e.filter {
+                    libc::EVFILT_READ | libc::EVFILT_SIGNAL => EventKind::Read,
+                    libc::EVFILT_WRITE => EventKind::Write,
+                    _ => unreachable!(),
+                };
+
                 Ok(TriggeredEvent {
                     index: e.udata as _,
                     data: e.data as _,
+                    kind,
                 })
             }
         })
@@ -168,8 +197,6 @@ impl Drop for Kqueue {
 }
 
 fn run_server(config: ApplicationConfig, listener: UnixListener) {
-    use io::Write;
-
     const NONE_PROCESS: Option<Process> = None;
 
     let mut application = match ServerApplication::new(config) {
@@ -178,10 +205,12 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
     };
 
     let mut client_connections: [Option<UnixStream>; MAX_CLIENT_COUNT] = Default::default();
+    let mut client_write_queue: [VecDeque<PooledBuf>; MAX_CLIENT_COUNT] = Default::default();
     let mut processes = [NONE_PROCESS; MAX_PROCESS_COUNT];
 
     let mut events = Vec::new();
     let mut timeout = None;
+    let mut need_redraw = false;
 
     const CLIENTS_START_INDEX: usize = 1;
     const CLIENTS_LAST_INDEX: usize = CLIENTS_START_INDEX + MAX_CLIENT_COUNT - 1;
@@ -189,16 +218,17 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
     const PROCESSES_LAST_INDEX: usize = PROCESSES_START_INDEX + MAX_PROCESS_COUNT - 1;
 
     let kqueue = Kqueue::new();
-    kqueue.add(Event::Fd(listener.as_raw_fd()), 0);
+    kqueue.add(Event::FdRead(listener.as_raw_fd()), 0, 0);
     let mut kqueue_events = KqueueEvents::new();
 
-    let _ignore_server_connection_buffer_len = ServerApplication::connection_buffer_len();
+    let _ignore_server_connection_buffer_len = SERVER_CONNECTION_BUFFER_LEN;
 
     loop {
+        let previous_timeout = timeout;
         let kqueue_events = kqueue.wait(&mut kqueue_events, timeout);
         if kqueue_events.len() == 0 {
             match timeout {
-                Some(Duration::ZERO) => timeout = Some(ServerApplication::idle_duration()),
+                Some(Duration::ZERO) => timeout = Some(SERVER_IDLE_DURATION),
                 Some(_) => {
                     events.push(PlatformEvent::Idle);
                     timeout = None;
@@ -210,9 +240,16 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
         }
 
         for event in kqueue_events {
-            let (event_index, event_data) = match event {
-                Ok(event) => (event.index, event.data),
-                Err(()) => return,
+            let (event_index, event_data, event_kind) = match event {
+                Ok(event) => (event.index, event.data, event.kind),
+                Err(()) => {
+                    for queue in &mut client_write_queue {
+                        for buf in queue.drain(..) {
+                            application.ctx.platform.buf_pool.release(buf);
+                        }
+                    }
+                    return;
+                }
             };
 
             match event_index {
@@ -220,14 +257,24 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                     for _ in 0..event_data {
                         match listener.accept() {
                             Ok((connection, _)) => {
+                                if let Err(error) = connection.set_nonblocking(true) {
+                                    panic!("could not set connection to nonblocking {}", error);
+                                }
+
                                 for (i, c) in client_connections.iter_mut().enumerate() {
                                     if c.is_none() {
                                         kqueue.add(
-                                            Event::Fd(connection.as_raw_fd()),
+                                            Event::FdRead(connection.as_raw_fd()),
                                             CLIENTS_START_INDEX + i,
+                                            libc::EV_CLEAR,
+                                        );
+                                        kqueue.add(
+                                            Event::FdWrite(connection.as_raw_fd()),
+                                            CLIENTS_START_INDEX + i,
+                                            libc::EV_CLEAR,
                                         );
                                         *c = Some(connection);
-                                        let handle = ClientHandle::from_index(i).unwrap();
+                                        let handle = ClientHandle(i as _);
                                         events.push(PlatformEvent::ConnectionOpen { handle });
                                         break;
                                     }
@@ -239,18 +286,41 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                 }
                 CLIENTS_START_INDEX..=CLIENTS_LAST_INDEX => {
                     let index = event_index - CLIENTS_START_INDEX;
+                    let handle = ClientHandle(index as _);
                     if let Some(ref mut connection) = client_connections[index] {
-                        let handle = ClientHandle::from_index(index).unwrap();
-                        match read_from_connection(
-                            connection,
-                            &mut application.ctx.platform.buf_pool,
-                            event_data as _,
-                        ) {
-                            Ok(buf) => events.push(PlatformEvent::ConnectionOutput { handle, buf }),
-                            Err(()) => {
-                                kqueue.remove(Event::Fd(connection.as_raw_fd()));
-                                client_connections[index] = None;
-                                events.push(PlatformEvent::ConnectionClose { handle });
+                        match event_kind {
+                            EventKind::Read => {
+                                match read_from_connection(
+                                    connection,
+                                    &mut application.ctx.platform.buf_pool,
+                                    event_data as _,
+                                ) {
+                                    Ok(buf) => {
+                                        events
+                                            .push(PlatformEvent::ConnectionOutput { handle, buf });
+                                    }
+                                    Err(()) => {
+                                        kqueue.remove(Event::FdRead(connection.as_raw_fd()));
+                                        kqueue.remove(Event::FdWrite(connection.as_raw_fd()));
+                                        client_connections[index] = None;
+                                        events.push(PlatformEvent::ConnectionClose { handle });
+                                    }
+                                }
+                            }
+                            EventKind::Write => {
+                                timeout = previous_timeout;
+
+                                let result = write_to_connection(
+                                    connection,
+                                    &mut application.ctx.platform.buf_pool,
+                                    &mut client_write_queue[index],
+                                );
+                                if result.is_err() {
+                                    kqueue.remove(Event::FdRead(connection.as_raw_fd()));
+                                    kqueue.remove(Event::FdWrite(connection.as_raw_fd()));
+                                    client_connections[index] = None;
+                                    events.push(PlatformEvent::ConnectionClose { handle });
+                                }
                             }
                         }
                     }
@@ -264,7 +334,7 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                             Ok(Some(buf)) => events.push(PlatformEvent::ProcessOutput { tag, buf }),
                             Err(()) => {
                                 if let Some(fd) = process.try_as_raw_fd() {
-                                    kqueue.remove(Event::Fd(fd));
+                                    kqueue.remove(Event::FdRead(fd));
                                 }
                                 process.kill();
                                 processes[index] = None;
@@ -277,36 +347,57 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
             }
         }
 
+        if events.is_empty() && !need_redraw {
+            continue;
+        }
+
+        need_redraw = false;
         application.update(events.drain(..));
         let mut requests = application.ctx.platform.requests.drain();
         while let Some(request) = requests.next() {
             match request {
                 PlatformRequest::Quit => {
-                    for request in requests {
-                        if let PlatformRequest::WriteToClient { buf, .. }
-                        | PlatformRequest::WriteToProcess { buf, .. } = request
-                        {
+                    for queue in &mut client_write_queue {
+                        for buf in queue.drain(..) {
                             application.ctx.platform.buf_pool.release(buf);
                         }
                     }
+                    for request in requests {
+                        drop_request(&mut application.ctx.platform.buf_pool, request);
+                    }
                     return;
                 }
-                PlatformRequest::Redraw => timeout = Some(Duration::ZERO),
+                PlatformRequest::Redraw => {
+                    need_redraw = true;
+                    timeout = Some(Duration::ZERO);
+                }
                 PlatformRequest::WriteToClient { handle, buf } => {
-                    let index = handle.into_index();
-                    if let Some(ref mut connection) = client_connections[index] {
-                        if connection.write_all(buf.as_bytes()).is_err() {
-                            kqueue.remove(Event::Fd(connection.as_raw_fd()));
-                            client_connections[index] = None;
-                            events.push(PlatformEvent::ConnectionClose { handle });
+                    let index = handle.0 as usize;
+                    match client_connections[index] {
+                        Some(ref mut connection) => {
+                            let write_queue = &mut client_write_queue[index];
+                            write_queue.push_back(buf);
+
+                            let result = write_to_connection(
+                                connection,
+                                &mut application.ctx.platform.buf_pool,
+                                write_queue,
+                            );
+                            if result.is_err() {
+                                kqueue.remove(Event::FdRead(connection.as_raw_fd()));
+                                kqueue.remove(Event::FdWrite(connection.as_raw_fd()));
+                                client_connections[index] = None;
+                                events.push(PlatformEvent::ConnectionClose { handle });
+                            }
                         }
+                        None => application.ctx.platform.buf_pool.release(buf),
                     }
-                    application.ctx.platform.buf_pool.release(buf);
                 }
                 PlatformRequest::CloseClient { handle } => {
-                    let index = handle.into_index();
+                    let index = handle.0 as usize;
                     if let Some(connection) = client_connections[index].take() {
-                        kqueue.remove(Event::Fd(connection.as_raw_fd()));
+                        kqueue.remove(Event::FdRead(connection.as_raw_fd()));
+                        kqueue.remove(Event::FdWrite(connection.as_raw_fd()));
                     }
                     events.push(PlatformEvent::ConnectionClose { handle });
                 }
@@ -325,7 +416,7 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                         if let Ok(child) = command.spawn() {
                             let process = Process::new(child, tag, buf_len);
                             if let Some(fd) = process.try_as_raw_fd() {
-                                kqueue.add(Event::Fd(fd), PROCESSES_START_INDEX + i);
+                                kqueue.add(Event::FdRead(fd), PROCESSES_START_INDEX + i, 0);
                             }
                             *p = Some(process);
                             events.push(PlatformEvent::ProcessSpawned { tag, handle });
@@ -342,7 +433,7 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                     if let Some(ref mut process) = processes[index] {
                         if !process.write(buf.as_bytes()) {
                             if let Some(fd) = process.try_as_raw_fd() {
-                                kqueue.remove(Event::Fd(fd));
+                                kqueue.remove(Event::FdRead(fd));
                             }
                             let tag = process.tag();
                             process.kill();
@@ -361,7 +452,7 @@ fn run_server(config: ApplicationConfig, listener: UnixListener) {
                     let index = handle.0 as usize;
                     if let Some(ref mut process) = processes[index] {
                         if let Some(fd) = process.try_as_raw_fd() {
-                            kqueue.remove(Event::Fd(fd));
+                            kqueue.remove(Event::FdRead(fd));
                         }
                         let tag = process.tag();
                         process.kill();
@@ -387,25 +478,26 @@ fn run_client(args: Args, mut connection: UnixStream) {
         Some(Terminal::new())
     };
 
-    let mut application = ClientApplication::new(terminal.as_ref().map(Terminal::to_file));
+    let mut application = ClientApplication::new();
+    application.output = terminal.as_ref().map(Terminal::to_client_output);
+
     let bytes = application.init(args);
     if connection.write_all(bytes).is_err() {
         return;
     }
 
     let kqueue = Kqueue::new();
-    kqueue.add(Event::Fd(connection.as_raw_fd()), 1);
+    kqueue.add(Event::FdRead(connection.as_raw_fd()), 1, 0);
     if is_pipped(libc::STDIN_FILENO) {
-        kqueue.add(Event::Fd(libc::STDIN_FILENO), 3);
+        kqueue.add(Event::FdRead(libc::STDIN_FILENO), 3, 0);
     }
 
     let mut kqueue_events = KqueueEvents::new();
 
     if let Some(terminal) = &terminal {
         terminal.enter_raw_mode();
-        kqueue.add(Event::Fd(terminal.as_raw_fd()), 0);
 
-        kqueue.add(Event::Resize, 2);
+        kqueue.add(Event::Resize, 2, 0);
 
         let size = terminal.get_size();
         let (_, bytes) = application.update(Some(size), &[Key::None], None, &[]);
@@ -422,29 +514,60 @@ fn run_client(args: Args, mut connection: UnixStream) {
     }
 
     let mut keys = Vec::new();
-    let buf_capacity =
-        ClientApplication::connection_buffer_len().max(ClientApplication::stdin_buffer_len());
+    let buf_capacity = CLIENT_CONNECTION_BUFFER_LEN.max(CLIENT_STDIN_BUFFER_LEN);
     let mut buf = Vec::with_capacity(buf_capacity);
 
+    let mut select_read_set = unsafe { std::mem::zeroed() };
+
     'main_loop: loop {
-        for event in kqueue.wait(&mut kqueue_events, None) {
+        keys.clear();
+
+        if let Some(terminal) = &terminal {
+            unsafe {
+                libc::FD_ZERO(&mut select_read_set);
+                libc::FD_SET(terminal.as_raw_fd(), &mut select_read_set);
+                libc::FD_SET(kqueue.as_raw_fd(), &mut select_read_set);
+
+                let result = libc::select(
+                    terminal.as_raw_fd().max(kqueue.as_raw_fd()) + 1,
+                    &mut select_read_set,
+                    std::ptr::null_mut(),
+                    std::ptr::null_mut(),
+                    std::ptr::null_mut(),
+                );
+                if result < 0 {
+                    break;
+                }
+
+                if libc::FD_ISSET(terminal.as_raw_fd(), &select_read_set) {
+                    buf.resize(buf_capacity, 0);
+                    match read(terminal.as_raw_fd(), &mut buf) {
+                        Ok(0) | Err(()) => break,
+                        Ok(len) => terminal.parse_keys(&buf[..len], &mut keys),
+                    }
+
+                    let (suspend, bytes) = application.update(None, &keys, None, &[]);
+                    if connection.write_all(bytes).is_err() {
+                        break;
+                    }
+                    if suspend {
+                        suspend_process(&mut application, Some(terminal));
+                    }
+
+                    if result == 1 {
+                        continue;
+                    }
+                }
+            }
+        }
+
+        for event in kqueue.wait(&mut kqueue_events, Some(Duration::ZERO)) {
             let mut resize = None;
             let mut stdin_bytes = None;
             let mut server_bytes = &[][..];
 
-            keys.clear();
-
             match event {
-                Ok(TriggeredEvent { index: 0, data }) => {
-                    if let Some(terminal) = &terminal {
-                        buf.resize(data as _, 0);
-                        match read(terminal.as_raw_fd(), &mut buf) {
-                            Ok(0) | Err(()) => break 'main_loop,
-                            Ok(len) => terminal.parse_keys(&buf[..len], &mut keys),
-                        }
-                    }
-                }
-                Ok(TriggeredEvent { index: 1, data }) => {
+                Ok(TriggeredEvent { index: 1, data, .. }) => {
                     buf.resize(data as _, 0);
                     match connection.read(&mut buf) {
                         Ok(0) | Err(_) => break 'main_loop,
@@ -454,11 +577,11 @@ fn run_client(args: Args, mut connection: UnixStream) {
                 Ok(TriggeredEvent { index: 2, .. }) => {
                     resize = terminal.as_ref().map(Terminal::get_size);
                 }
-                Ok(TriggeredEvent { index: 3, data }) => {
+                Ok(TriggeredEvent { index: 3, data, .. }) => {
                     buf.resize(data as _, 0);
                     match read(libc::STDIN_FILENO, &mut buf) {
                         Ok(0) | Err(()) => {
-                            kqueue.remove(Event::Fd(libc::STDIN_FILENO));
+                            kqueue.remove(Event::FdRead(libc::STDIN_FILENO));
                             stdin_bytes = Some(&[][..]);
                         }
                         Ok(len) => stdin_bytes = Some(&buf[..len]),
@@ -473,7 +596,7 @@ fn run_client(args: Args, mut connection: UnixStream) {
                 break;
             }
             if suspend {
-                suspend_process(&mut application, &terminal);
+                suspend_process(&mut application, terminal.as_ref());
             }
         }
     }
