@@ -1,9 +1,9 @@
 use std::{
+    collections::VecDeque,
     env, fs, io,
-    mem::ManuallyDrop,
     os::unix::{
         ffi::OsStrExt,
-        io::{AsRawFd, FromRawFd, RawFd},
+        io::{AsRawFd, RawFd},
         net::{UnixListener, UnixStream},
     },
     path::Path,
@@ -73,6 +73,7 @@ pub(crate) fn run(
             Err(_) => match unsafe { libc::fork() } {
                 -1 => panic!("could not start server"),
                 0 => {
+                    unsafe { libc::daemon(true as _, false as _) };
                     server_fn(config, start_server(session_path));
                     let _ = fs::remove_file(session_path);
                 }
@@ -115,8 +116,8 @@ impl Terminal {
         Self { fd, original_state }
     }
 
-    pub fn to_file(&self) -> ManuallyDrop<fs::File> {
-        unsafe { ManuallyDrop::new(fs::File::from_raw_fd(self.fd)) }
+    pub fn to_client_output(&self) -> ClientOutput {
+        ClientOutput(self.fd)
     }
 
     pub fn enter_raw_mode(&self) {
@@ -143,7 +144,7 @@ impl Terminal {
         unsafe { libc::tcsetattr(self.fd, libc::TCSAFLUSH, &self.original_state) };
     }
 
-    pub fn get_size(&self) -> (usize, usize) {
+    pub fn get_size(&self) -> (u16, u16) {
         let mut size: libc::winsize = unsafe { std::mem::zeroed() };
         let result = unsafe {
             libc::ioctl(
@@ -248,15 +249,71 @@ pub(crate) fn read_from_connection(
 ) -> Result<PooledBuf, ()> {
     use io::Read;
     let mut buf = buf_pool.acquire();
-    let write = buf.write_with_len(len);
-    match connection.read(write) {
-        Ok(0) | Err(_) => {
-            buf_pool.release(buf);
-            Err(())
+    let write = buf.write();
+
+    loop {
+        let start = write.len();
+        write.resize(start + len, 0);
+        match connection.read(&mut write[start..start + len]) {
+            Err(error) => {
+                match error.kind() {
+                    io::ErrorKind::WouldBlock => write.truncate(start),
+                    _ => write.clear(),
+                }
+                break;
+            }
+            Ok(len) => {
+                write.truncate(start + len);
+                if len == 0 {
+                    break;
+                }
+            }
         }
-        Ok(len) => {
-            write.truncate(len);
-            Ok(buf)
+    }
+
+    if write.is_empty() {
+        buf_pool.release(buf);
+        Err(())
+    } else {
+        Ok(buf)
+    }
+}
+
+pub(crate) fn write_to_connection(
+    connection: &mut UnixStream,
+    buf_pool: &mut BufPool,
+    write_queue: &mut VecDeque<PooledBuf>,
+) -> Result<(), ()> {
+    use io::Write;
+
+    loop {
+        let mut buf = match write_queue.pop_front() {
+            Some(buf) => buf,
+            None => return Ok(()),
+        };
+
+        match connection.write(buf.as_bytes()) {
+            Ok(len) => {
+                buf.drain_start(len);
+                if buf.as_bytes().is_empty() {
+                    buf_pool.release(buf);
+                } else {
+                    write_queue.push_front(buf);
+                }
+            }
+            Err(error) => match error.kind() {
+                io::ErrorKind::WouldBlock => {
+                    write_queue.push_front(buf);
+                    return Ok(());
+                }
+                _ => {
+                    buf_pool.release(buf);
+                    for buf in write_queue.drain(..) {
+                        buf_pool.release(buf);
+                    }
+                    return Err(());
+                }
+            },
         }
     }
 }
@@ -335,7 +392,12 @@ impl Drop for Process {
     }
 }
 
-pub(crate) fn suspend_process(application: &mut ClientApplication, terminal: &Option<Terminal>) {
+pub(crate) fn suspend_process<O>(
+    application: &mut ClientApplication<O>,
+    terminal: Option<&Terminal>,
+) where
+    O: io::Write,
+{
     application.restore_screen();
     if let Some(terminal) = terminal {
         terminal.leave_raw_mode();
@@ -348,3 +410,19 @@ pub(crate) fn suspend_process(application: &mut ClientApplication, terminal: &Op
     }
     application.reinit_screen();
 }
+
+pub struct ClientOutput(RawFd);
+impl io::Write for ClientOutput {
+    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
+        let len = unsafe { libc::write(self.0, buf.as_ptr() as _, buf.len()) };
+        if len >= 0 {
+            Ok(len as _)
+        } else {
+            Err(io::Error::last_os_error())
+        }
+    }
+
+    fn flush(&mut self) -> io::Result<()> {
+        Ok(())
+    }
+}
